@@ -17,7 +17,7 @@ import unittest
 import gnupg
 
 __author__ = "Vinay Sajip"
-__date__  = "$07-Dec-2014 18:46:40$"
+__date__  = "$24-Sep-2015 18:05:15$"
 
 ALL_TESTS = True
 
@@ -230,6 +230,13 @@ class GPGTestCase(unittest.TestCase):
         public_keys = self.gpg.list_keys()
         self.assertTrue(is_list_with_len(public_keys, 1),
                         "1-element list expected")
+        key_info = public_keys[0]
+        fp = key_info['fingerprint']
+        self.assertTrue(fp in public_keys.key_map)
+        self.assertTrue(public_keys.key_map[fp] is key_info)
+        for _, _, sfp in key_info['subkeys']:
+            self.assertTrue(sfp in public_keys.key_map)
+            self.assertTrue(public_keys.key_map[sfp] is key_info)
         private_keys = self.gpg.list_keys(secret=True)
         self.assertTrue(is_list_with_len(private_keys, 1),
                         "1-element list expected")
@@ -243,6 +250,39 @@ class GPGTestCase(unittest.TestCase):
         private_keys_2 = gpg.list_keys(secret=True)
         self.assertEqual(private_keys_2, private_keys)
 
+        # generate additional keys so that we can test listing a subset of
+        # keys
+        def get_names(key_map):
+            result = set()
+            for info in key_map.values():
+                for uid in info['uids']:
+                    uid = uid.replace(' (A test user (insecure!))', '')
+                    result.add(uid)
+            return result
+
+        result = self.generate_key("Charlie", "Clark", "gamma.com")
+        self.assertNotEqual(None, result, "Non-null result")
+        result = self.generate_key("Donna", "Davis", "delta.com")
+        self.assertNotEqual(None, result, "Non-null result")
+        public_keys = gpg.list_keys()
+        self.assertEqual(len(public_keys), 3)
+        actual = get_names(public_keys.key_map)
+        expected = set(['Barbara Brown <barbara.brown@beta.com>',
+                        'Charlie Clark <charlie.clark@gamma.com>',
+                        'Donna Davis <donna.davis@delta.com>'])
+        self.assertEqual(actual, expected)
+        # specify a single key as a string
+        public_keys = gpg.list_keys(keys='Donna Davis')
+        actual = get_names(public_keys.key_map)
+        expected = set(['Donna Davis <donna.davis@delta.com>'])
+        self.assertEqual(actual, expected)
+        # specify multiple keys
+        public_keys = gpg.list_keys(keys=['Donna', 'Barbara'])
+        actual = get_names(public_keys.key_map)
+        expected = set(['Barbara Brown <barbara.brown@beta.com>',
+                        'Donna Davis <donna.davis@delta.com>'])
+        self.assertEqual(actual, expected)
+
     def test_scan_keys(self):
         "Test that external key files can be scanned"
         expected = set([
@@ -515,6 +555,16 @@ class GPGTestCase(unittest.TestCase):
                 logger.debug("was: %r", data)
                 logger.debug("new: %r", ddata)
             self.assertEqual(data, ddata, "Round-trip must work")
+
+            # Try opening the encrypted file in text mode (Issue #39)
+            # this doesn't fail in 2.x
+            if gnupg._py3k:
+                efile = open(encfname, 'r')
+                ddata = self.gpg.decrypt_file(efile, passphrase="bbrown",
+                                              output=decfname)
+                self.assertFalse(ddata)
+                self.assertEqual(ddata.status, "no data was provided")
+                efile.close()
         finally:
             for fn in (encfname, decfname):
                 if os.name == 'posix' and mode is not None:
